import sys, os, shutil, re

from const import *

# WineDoors global instances
from preferences import preferences
from packlist import packlist
from utils import Download
from utils import printNestedStuff
from wine import wine
from log import log

# Parser
from xml.sax import make_parser
from xml.sax.handler import ContentHandler

class ApplicationPackParse( ContentHandler ):
    def __init__( self, pack_data ):
        self.in_app_pack = False
        ContentHandler.__init__( self )
        self.pack_data = pack_data
        self.requires = {}
        self.data = []
        self.reg = False
        self.desktop = False
        self.pack_data['winereg'] = None
        self.pack_data['desktop'] = None
        self.pack_data['major-version'] = None
        
    def startElement( self, name, attrs ):
        if name == "ApplicationPack":
            self.in_app_pack = True
        elif name == "crack" and self.in_app_pack:
            self.crack_required = attrs.get( 'required' ) == "1"
        elif name == "requires" and self.in_app_pack:
            self.requires[attrs.get( 'name' )] = attrs.get( 'version' )
            
    def endElement( self, name ):
        if name == "ApplicationPack":
            self.in_app_pack = False
            self.pack_data['requires'] = self.requires
        elif name == "crack" and self.in_app_pack:
            self.pack_data["crack_required"] = self.crack_required
            text = ''.join( self.data )
            self.pack_data[name] = text
        elif self.in_app_pack:
            text = ''.join( self.data )
            self.pack_data[name] = text
        self.data = []

    def characters( self, string ):
        self.data.append( string.rstrip() )
            
    def error( self, exception ):
        log.Error( "%s\n" % exception )

class ApplicationPack:
    def __init__( self, operation, pack, version=None, repo=None ):
        """
        pack can be a remote file, local file or pack short name.
        pack short names require repo to be set. If no version is specified
        default version will be installed, if no default version is specified
        all versions will be installed.
        """
        self.app_pack = {}
        self.pack = []
        self.repo = None
        self.requires = []
        self.version = version
        self.operation = operation
        self.repo = repo
        self.repo_uri = None
        
        if pack.startswith ( "http://" ): 
            self.pack_type = REMOTE
            self.pack_file = pack
        elif os.path.isfile( pack ):
            self.pack_type = LOCAL
            self.pack_file = pack
        elif self.repo or packlist.InAvailable( pack ):
            self.pack = packlist.GetPack(pack)
            
            if not self.version:
                try:
                    log.Debug("Setting default version: "+ self.pack['default_version'])
                    self.version = self.pack['default_version']
                except:
                    log.Debug("Installing all versions")
                    self.version = ALL
            
            if self.repo:
                self.repo_uri = packlist.GetRepoURI( repo )
                self.pack_type = REPO
            else:
                repo_data = packlist.GetRepo( pack )
                self.repo = ''.join( repo_data[0:1] )
                self.repo_uri = ''.join( repo_data[1:] )
                self.pack_type = PACKLIST
            
            if self.version != ALL:
                for pack_version in self.pack['versions']:
                    [( majorversion, 
                       wineversion, 
                       winevariant, 
                       appdb_rating, 
                       installersize, 
                       installedsize, 
                       votes,
    		           version,
                       app_pack )] = pack_version
                    
                    if majorversion == self.version:
                        self.pack_file = app_pack
                        break
        else:
            log.Error( "Cannot determine the application pack type for " + pack )
            print repo
            return
        
        if self.operation == INSTALL or operation == DRYRUN or operation == UPDATE:
            log.Debug( "Installing application " )
            self.InstallApplication()
        elif self.operation == UNINSTALL:
            self.repo = "installed"
            self.UninstallApplication()
        
    def Parse( self ):
        self.app_pack_xml = self.pack_path + "pack.xml"
        log.Debug("Parsing file: " + self.app_pack_xml )
        handler = ApplicationPackParse( self.app_pack )
        parser = make_parser()
        parser.setContentHandler( handler )
        try:
            parser.parse( open( self.app_pack_xml ) )
        except:
            parser.parse( self.app_pack_xml )

    def Download ( self, pack_file, file_type=APP_PACK ):
        log.Debug("Downloading file: " + pack_file)
        
        # Marcelo in the blame file again?!
        # This invalidates the download logic see ticket #458
        #if not self.repo:
        #    self.repo = ""
        #log.Debug("Repo: " +self.repo)
        
        if pack_file.startswith("http://") or pack_file.startswith("ftp://") or not self.repo:
            remote_file = pack_file
            ( directory, file_name ) = os.path.split( remote_file )
        elif self.repo:
            log.Debug("Repo URI: "+ self.repo_uri)
            remote_file = self.repo_uri + pack_file
            file_name = pack_file
        
        self.cache_path = os.environ['HOME'] + "/.wine-doors/apppacks/"
        
        if not os.path.isdir(self.cache_path):
            os.makedirs( self.cache_path )
        
        if file_type == APP_PACK:
            if file_name.endswith(".tar.gz"):
                self.pack_folder = file_name.replace( ".tar.gz", "" )
            elif file_name.endswith(".tar.bz2"):
                self.pack_folder = file_name.replace( ".tar.bz2", "" )
            else:
                ( self.pack_folder, ext ) = os.path.splitext( file_name )

            self.pack_path = self.cache_path + self.pack_folder + "/"
            self.cache_file = self.cache_path + file_name
        elif file_type == WINDOWS_EXE:
            self.cache_file = self.pack_path + "resources/" + file_name
            if not os.path.isdir( self.pack_path + "resources" ):
                os.makedirs( self.pack_path + "resources" )
                
        log.Debug( "Remote file: " + remote_file )
        log.Debug( "Pack path: " + self.pack_path )
        log.Debug( "Cache file: " + self.cache_file )
        
        Download ( remote_file, self.cache_file, True )
        self.cache_file = self.cache_file.replace( ".gz", "" )

    def Extract ( self ):
        cwd = os.getcwd()
        os.chdir(self.cache_path)
        log.Debug("Extracting tarball:" + self.cache_file)
        if self.cache_file.endswith( "bz2" ):
            os.system( "tar -xjf " + self.cache_file )
        elif self.cache_file.endswith( "tar" ):
            os.system( "tar -xf " + self.cache_file )
        elif self.cache_file.endswith( "wdi" ):
            os.system( "tar -xzf " + self.cache_file )
        os.chdir(cwd)
                
    def InstallApplication( self ):
        """
        Install an application using an application pack
        """
        if self.version == ALL:
            for pack_version in self.pack['versions']:
                [( majorversion, 
                   wineversion, 
                   winevariant, 
                   appdb_rating, 
                   installersize, 
                   installedsize, 
                   votes,
                   version,
                   app_pack )] = pack_version
                log.Debug("Forking install for version: " + majorversion )
                self.pack_file = app_pack
                self.version = majorversion
                self.InstallApplication()
        else:
            if self.operation == DRYRUN:
                log.Debug("Doing dry run install")
            
            log.Debug("* Retrieving file to cache: " + self.pack_file)
            
            self.Download( self.pack_file, APP_PACK )
            self.Extract()
            self.Parse()
            try:
                if self.app_pack['url'] != "":
                    log.Debug("* Retrieving installer executable to cache: " + self.app_pack['url'])
                    self.Download( self.app_pack['url'], WINDOWS_EXE )
            except:
                pass
            
            try:
                if re.search("D:\\\\", self.app_pack['install']):
                    # Its most likely a CD-ROM
                    installer = self.app_pack['install']
                elif re.search("CDROM:/", self.app_pack['install']):
                    # Erm what shall we do here?
                    pass
                else:
                    installer = self.pack_path + self.app_pack['install']
            except:
                installer = self.cache_file
                                
            try:
                installer_args = self.app_pack['installer_args']
            except:
                installer_args = ""
            
            if self.operation != DRYRUN:
                cwd = os.getcwd()
                os.chdir( self.pack_path )
                new_cwd = os.getcwd()
                # Execute installer
                log.Debug("* About to execute: "+installer+" "+installer_args )
                log.Debug("** From folder: " + new_cwd )
                self.status = wine.Execute( installer, installer_args )
                os.chdir(cwd)
                if wine.WaitForExit() == False:
                    wine.Kill()
                if self.status == FILE_NOT_FOUND:
                    log.Error("Can't find file: " + installer)
                    #browse_file = ui.CantFindFile()
                    #installer = browse_file.get_file()
                    self.status = wine.Execute( installer, installer_args )
                    if wine.WaitForExit() == False:
                        wine.Kill()
            if self.app_pack['desktop'] and self.app_pack['desktop'] != "":
                self.InstallFile( self.pack_path + self.app_pack['desktop'], LAUNCHER )
            if self.app_pack['winereg'] and self.app_pack['winereg'] != "":
                self.InstallFile( self.pack_path + self.app_pack['winereg'], REGISTRY )
            
            # This is here just in case the XML hasn't been updated to include
            # the major-version tag
            if not self.app_pack['major-version']:
                majorversion = self.app_pack['version']
            else:
                majorversion = self.app_pack['major-version']
            
            if not self.status:
                packlist.AddInstalled(self.app_pack['shortname'], majorversion, self.app_pack['version'])
            return False

    def InstallFile( self, resource_file, resource_type ):
        if resource_type == LAUNCHER:
            desktop_path = os.path.expanduser( "~/.local/share/applications/" )
            if not os.path.isdir( desktop_path ):
                os.makedirs( desktop_path )
            if re.search(":\\\\", self.app_pack['run'] ):
                quote_exe = re.compile(".exe( )", re.IGNORECASE)
                exe = quote_exe.sub("\" ", self.app_pack['run'])
                run = "env WINEPREFIX="+preferences['wineroot']+" wine \""+exe 
            else:
                ( directory, runfile ) = os.path.split( self.app_pack['run'] )
                shutil.copy(self.app_pack['run'], os.path.expanduser( "~/bin/"))
                run = os.path.expanduser( "~/bin/" + runfile )
                os.system("chmod +x "+run )
            desktop = open( resource_file , 'a' )
            desktop.write( 'Exec=' + run+"\n" )
            ( directory, icon_file ) = os.path.split( self.app_pack['icon'] )
            desktop.write( 'Icon=' + desktop_path + icon_file+"\n" )
            desktop.close()
            shutil.copy( resource_file, desktop_path )
            self.InstallFile( self.pack_path + self.app_pack['icon'], ICON )
        elif resource_type == ICON:
            desktop_path = os.path.expanduser( "~/.local/share/applications/" )
            if not os.path.isdir( desktop_path ):
                os.makedirs( desktop_path )
            shutil.copy( resource_file, desktop_path )
        elif resource_type == SCRIPT:
            if not os.path.isdir( os.environ['HOME']+"/bin/" ):
                os.makedirs( os.environ['HOME']+"/bin/" )
            shutil.copy( resource_file, os.environ['HOME']+"/bin/" )
        elif resource_type == REGISTRY:
            wine.Execute( "regedit", resource_file )
    
    def UninstallApplication(self):
        log.Debug("Uninstalling application: " + self.pack + " " + version)
            
''' Unit testing '''
def usage():
    print "Unit testing of pack.py"
    print "Usage: python pack.py <class> <file>"
    print "  <class> one of ApplicationPackParse, ApplicationPack"
    print "  <file> file to feed into the class as a unit"
    
if __name__ == "__main__":
    if len( sys.argv ) == 3:
        if sys.argv[1] == "ApplicationPack":
            print "Testing Class: ApplicationPack"
            print "  Parsing application pack from " + sys.argv[2]            
            apppack = ApplicationPack( INSTALL, sys.argv[2] )
        elif sys.argv[1] == "ApplicationPackParse":
            print "Testing Class: ApplicationPackParse"
            print "  Parsing application pack xml from " + sys.argv[2]
            pack = []
            handler = ApplicationPackParse( pack )
            parser = make_parser()
            parser.setContentHandler( handler )
            parser.parse( sys.argv[2] )
            printNestedStuff( pack )   
        else:
            print sys.argv[1] + " is not a valid class"
            usage()       
    else:
        usage()
